﻿using System;
using System.ComponentModel;
using System.Collections.ObjectModel;
using XiquiparcsReunits_Silverlight.Tools;
using System.Collections.Generic;
using System.Windows;
using XiquiparcsReunits_Silverlight.GestorXiquiparc;
using System.ServiceModel;
using System.ServiceModel.Channels;

namespace XiquiparcsReunits_Silverlight.Model
{
    /// <summary>
    /// Objecte que registra els bens o serveis oferts a un client en un periode determinat. Es tanca la facturació, un cop
    /// el client ha pagat la totalitat de l'import debengat. En aquest estat no es podrà modificar.
    /// També registra reserves futures, ja que es factures, i passen a estat de cobrament pendent.
    /// </summary>
    public class Factura : INotifyPropertyChanged
    {
        private string _id_factura;
        private string _estat_cobrament;
        private DateTime _data_cobrament;
        private string _descripcio;
        private Decimal _a_compte;
        private bool _modificable;

        /// <summary>
        /// Identificador sobre si aquesta factura té recursos facturats al futur,
        /// o sigui, que té reserves de recursos.
        /// </summary>
        private bool esReserva = false;
        public bool EsReserva
        {
            get
            {
                return esReserva;
            }
            set
            {
                esReserva = value; OnPropertyChanged("EsReserva");
            }
        }

        /// <summary>
        /// Client al qual es factura
        /// </summary>
        private string id_usuari_factura;



        /// <summary>
        /// Client al qual es factura
        /// </summary>
        public string Id_usuari_factura { get { return id_usuari_factura; } set { id_usuari_factura = value; OnPropertyChanged(Id_usuari_factura); } }
        private Usuari usuari_factura;


        /// <summary>
        /// Client al qual es factura
        /// </summary>
        public Usuari Usuari_factura
        {
            get
            {
                return usuari_factura;
            }
            set
            {
                Id_usuari_factura = value.Id_usuari;
                usuari_factura = value;
                OnPropertyChanged("Usuari_factura");
            }
        }

        /// <summary>
        /// Identificador de la factura actual
        /// </summary>
        public string Id_factura
        {
            get { return _id_factura; }
            set
            {
                _id_factura = value;
                OnPropertyChanged("Id_factura");
                OnPropertyChanged("FacturaCreada");
            }
        }

        /// <summary>
        /// Estat de cobrament de la factura
        /// </summary>
        public string Estat_cobrament
        {
            get { return _estat_cobrament; }
            set
            {
                _estat_cobrament = value;
                OnPropertyChanged("Estat_cobrament");
            }
        }

        /// <summary>
        /// Data en que s'ha fet el darrer cobrament
        /// </summary>
        public DateTime Data_cobrament
        {
            get { return _data_cobrament; }
            set
            {
                _data_cobrament = value;
                OnPropertyChanged("Data_cobrament");
            }
        }

        /// <summary>
        /// Descripció complementaria de la factura
        /// </summary>
        public string Descripcio
        {
            get { return _descripcio; }
            set
            {
                _descripcio = value;
                OnPropertyChanged("Descripcio");
            }
        }

        /// <summary>
        /// Booleà que mostra si la factura està en estat modificable o no
        /// </summary>
        public bool Modificable
        {
            get { return _modificable; }
            set
            {
                _modificable = value;
                OnPropertyChanged("Modificable");
            }
        }

        /// <summary>
        /// Contador del nombre de recursos facturats a la factura actual
        /// </summary>
        public int contador_recursos
        {
            get
            {
                return Recursos_facturats.Count;
            }
        }

        /// <summary>
        /// Llista de tots els recursos facturats en l'actual factura
        /// </summary>
        private ObservableCollection<Recurs_facturat> recursos;
        public ObservableCollection<Recurs_facturat> Recursos_facturats
        {
            get
            {
                return recursos;
            }
            set
            {
                recursos = value;
                OnPropertyChanged("contador_recursos");
                OnPropertyChanged("Recursos");
                OnPropertyChanged("Import_total");
                OnPropertyChanged("Pendent");
                OnPropertyChanged("Impostos");
                OnPropertyChanged("Base_imponible");
            }
        }

        /// <summary>
        /// Import total dels aportaments a compte que ha fet l'usuari
        /// </summary>
        public Decimal A_compte
        {
            get { return _a_compte; }
            set
            {
                _a_compte = value;
                OnPropertyChanged("A_compte");
                OnPropertyChanged("Pendent");
            }
        }

        /// <summary>
        /// Import pendent de pagar pel client, diferencia entre import total i a compte
        /// </summary>
        public Decimal Pendent
        {
            get
            {
                return Import_total - A_compte;
            }
        }

        /// <summary>
        /// Import total de la factura, sense impostos
        /// </summary>
        public Decimal Base_imponible
        {
            get
            {
                Decimal import = 0;
                foreach (Recurs_facturat recurs_facturat in this.Recursos_facturats)
                    import = import + (recurs_facturat.Import * recurs_facturat.Quantitat);
                return import;
            }
        }

        /// <summary>
        /// Import total dels impostos
        /// </summary>
        public Decimal Impostos
        {
            get { return Import_total - Base_imponible; }
        }

        /// <summary>
        /// Import total de la factura, incloient IVA
        /// </summary>
        public Decimal Import_total
        {
            get
            {
                Decimal import = 0;
                foreach (Recurs_facturat recurs_facturat in this.Recursos_facturats)
                    import = import + (recurs_facturat.Import * recurs_facturat.Quantitat * (1 + recurs_facturat.IVA_recurs));
                return import;
            }
        }

        /// <summary>
        /// Propietat que representa si una factura es modificable i té identificador,
        /// o sigui, si se li poden afegir recursos facturats
        /// </summary>
        #region ViewModelProperty: FacturaCreada
        public bool FacturaCreada
        {
            get
            {
                return Modificable && Id_factura != null;
            }
        }
        #endregion

        /// <summary>
        /// Constructor de la factura sense paràmetres que NO inicialitza els recursos facturats
        /// </summary>
        private Factura()
        {
            Recursos_facturats = new ObservableCollection<Recurs_facturat>();
        }
        
        public event PropertyChangedEventHandler PropertyChanged;

        public void OnPropertyChanged(string name)
        {
            if (PropertyChanged != null)
                PropertyChanged(this, new PropertyChangedEventArgs(name));
        }

        /// <summary>
        /// Funció que retorna una factura inicialitzant els mínims paràmetres
        /// </summary>
        /// <param name="id_factura">Identificador de la factura (no es comprova)</param>
        /// <param name="id_usuari_factura">Identificador de l'usuari (no es comprova)</param>
        /// <param name="estat_cobrament">Identifciador de l'estat de cobrament (no es comprova)</param>
        /// <param name="modificable">Estat de la factura, que permet modificar-la</param>
        /// <param name="a_compte">Import que l'usuari a donat a compte</param>
        /// <param name="data_cobrament">Data en que s'ha produït el darrer cobrament</param>
        /// <param name="descripcio">Descripcio de la factura</param>
        /// <returns>Factura creada</returns>
        public static Factura CreateFactura(String id_factura, String id_usuari_factura, String estat_cobrament, bool modificable, decimal a_compte, DateTime data_cobrament, String descripcio)
        {
            Factura factura = new Factura();
            factura.Id_factura = id_factura;
            factura.Id_usuari_factura = id_usuari_factura;
            factura.Estat_cobrament = estat_cobrament;
            factura.Modificable = modificable;
            factura.A_compte = a_compte;
            factura.Descripcio = descripcio;
            factura.Data_cobrament = data_cobrament;
            return factura;
        }

        /// <summary>
        /// Funció que retorna un identificador de la factura
        /// </summary>
        /// <returns>Cadena que representa a una factura</returns>
        public override string ToString()
        {
            return this.Id_factura;
        }
    }
}